import queue
import time
import itertools
import selectors
import logging
import threading
from .eventfd import eventfd
from concurrent.futures import ThreadPoolExecutor
from .utils import create_thread_pool, sockinfo, log, LoggerAdapter, flag_to_str
from .channel import ChannelFuture, AbstractChannel
from typing import List, Tuple
import os
import inspect
from attrs import define, field

logger = LoggerAdapter(logging.getLogger(__name__))

DEBUG_INTERVAL_MILLIS = int(os.getenv('PY_NETTY_DEBUG_INTERVAL_MILLIS', 60000))


class EventLoop:

    def __init__(self, pool: ThreadPoolExecutor):
        assert pool, "thread pool executor is required"

        # internals
        self._channels = {}  # {fileno: Channel}
        self._connect_timeout_due_millis = {}  # {fileno: due_millis}
        self._thread = None
        self._stop_polling = False
        self._start_barrier = threading.Event()
        self._lock = threading.Lock()
        self._pool = pool

        # create selector
        self._eventfd = eventfd()
        self._selector = selectors.DefaultSelector()
        self._selector.register(self._eventfd, selectors.EVENT_READ)
        logger.debug("selector(%s) created for pool [%s]", type(self._selector).__name__, self._pool._thread_name_prefix)

        # queues
        self._taskq = queue.Queue()

        # counters
        self._eventfd_read_count = 0
        self._eventfd_write_count = 0
        self._total_accepted = 0
        self._total_sent = 0
        self._total_received = 0
        self._total_registered = 0
        self._total_tasks_submitted = 0
        self._total_tasks_processed = 0

    def modify_flag(self, channel):
        if not self.in_eventloop():
            self.submit_task(lambda: self.modify_flag(channel))
            return

        fileno = channel._fileno
        flag = channel._flag

        if flag == 0:
            self._selector.unregister(fileno)
            return

        try:
            self._selector.get_key(fileno)
        except KeyError:
            if channel.socket().fileno() == -1:
                return
            self._selector.register(channel, flag)
        else:
            self._selector.modify(fileno, flag)

    def submit_task(self, task):
        self.start()
        self._taskq.put(task)
        self._total_tasks_submitted += 1
        self.interrupt("submit task")

    def interrupt(self, desc=""):
        if desc and logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"interrupting eventloop with EventFD {hex(id(self._eventfd))} in {self._thread.name}: {desc}")
        self._eventfd.unsafe_write()

        if not logger.isEnabledFor(logging.DEBUG):
            return
        # only in debug mode to accumulate counter
        with self._lock:
            self._eventfd_write_count += 1

    def unregister(self, channel: AbstractChannel, channel_future: ChannelFuture = None):
        cf = channel_future or ChannelFuture(channel)
        if not self.in_eventloop():
            self.submit_task(lambda: self.unregister(channel, cf))
            return cf
        fileno = channel.fileno0()
        try:
            self._selector.unregister(fileno)
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("unregistered channel %s/%s from selector", channel.id(), fileno)
            channel.handler_context().fire_channel_unregistered()
        except Exception:
            pass
        self._channels.pop(fileno, None)
        self._connect_timeout_due_millis.pop(fileno, None)
        cf.set(channel)
        return cf

    def in_eventloop(self):
        return self._thread == threading.current_thread()

    def register(self, channel: AbstractChannel, only_write=False) -> ChannelFuture:
        self.start()

        if not self.in_eventloop():
            self.submit_task(lambda: self.register(channel, only_write))
            return channel.channel_future()

        if channel.socket().fileno() == -1:
            # channel closed already
            return channel.channel_future()

        channel.socket().setblocking(False)

        if only_write:
            flag = selectors.EVENT_WRITE
        else:
            flag = selectors.EVENT_READ | selectors.EVENT_WRITE
        channel.set_flag(flag)
        self._selector.register(channel, flag)
        channel.handler_context().fire_channel_registered()
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("registered channel(server:%s) [%s/%s] with flag: %s(%s)",
                         channel.is_server(), channel.id(), channel.fileno(), flag, flag_to_str(flag))
        self._total_registered += 1
        self._channels[channel.fileno()] = channel
        if not channel.is_server():
            self._connect_timeout_due_millis[channel.fileno()] = int(time.time() * 1000) + channel.connect_timeout_millis()

        # cf.set(channel)
        return channel.channel_future()

    def stop(self):
        logger.debug("stopping poll")
        self._stop_polling = True
        self.interrupt('stop poll')

    def _process_task_queue(self):
        while not self._taskq.empty():
            task = self._taskq.get()
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("task to run: %s", task)
            start = time.time()
            try:
                task()
            except Exception:
                logger.exception("error when running task: \n%s", inspect.getsource(task))
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("task finished in %sms: \n%s", int((time.time() - start) * 1000), inspect.getsource(task))
            self._total_tasks_processed += 1

    def _close_channel_internally(self, channel, reason=''):
        assert self.in_eventloop(), "Must be in event loop"
        logger.debug(f"closing channel internally (reason: {reason}): {channel}")
        channel.socket().close()
        channel.close_future().set(channel)
        channel.set_active(False, reason)
        channel.unregister()

    def _events_to_str(self, events: List[Tuple[selectors.SelectorKey, int]]):
        result = []
        for key, flag in events:
            fileno = key.fd
            if fileno == self._eventfd.fileno():
                fd_name = f"EventFD({hex(id(self._eventfd))})"
            else:
                channel = self._channels.get(fileno)
                if not channel:
                    fd_name = f"unknown({fileno})"
                else:
                    fd_name = "%s(%s/%s)" % ('server' if channel.is_server() else 'client', fileno, channel.id())
            flags_str = flag_to_str(flag)
            result.append(f"{fd_name}:{flags_str}")
        return ", ".join(result)

    def _show_debug_info(self, n=50):
        # logger.debug(f'{"=" * n} {threading.current_thread().name} {"=" * n}')
        logger.debug(" counters ".center(n, '='))
        logger.debug("eventfd writes:        %s", self._eventfd_write_count)
        logger.debug("eventfd reads:         %s", self._eventfd_read_count)
        logger.debug("pending tasks:         %s", self._taskq.qsize())
        logger.debug("total sent bytes:      %s", self._total_sent)
        logger.debug("total received:        %s", self._total_received)
        logger.debug("total registered:      %s", self._total_registered)
        logger.debug("total accepted:        %s", self._total_accepted)
        logger.debug("total tasks submitted: %s", self._total_tasks_submitted)
        logger.debug("total tasks processed: %s", self._total_tasks_processed)
        logger.debug("pending connections:   %s", len(self._connect_timeout_due_millis))
        logger.debug("active connections:    %s", max(0, len(self._channels) - len(self._connect_timeout_due_millis)))
        logger.debug("current hooked:        %s", len(self._selector.get_map()))
        logger.debug("hooks:                 %s", self._selector._fd_to_key)

        logger.debug(" channels ".center(n, '='))
        for channel in self._channels.values():
            logger.debug(f"{channel}")

        logger.debug(" pendings ".center(n, '='))
        for channel in self._channels.values():
            if channel.is_server():  # server channel has no pendings
                continue
            channel_id = channel.id()
            if not channel.has_pendings():
                # logger.debug(f"{channel_id}: no pendings")
                continue
            chunk_count = 0
            bytes_count = 0
            for chunk in channel.pendings():
                chunk_count += 1
                bytes_count += len(chunk.buffer)
            logger.debug(f"{channel_id}: {chunk_count} chunks, {bytes_count} bytes in total")

    def _millis_to_wait_for_connect_timeout(self) -> int:  # in milliseconds
        if not self._connect_timeout_due_millis:
            return -1  # wait forever
        min_timeout = min(self._connect_timeout_due_millis.values())  # nearest timeout
        return max(0, min_timeout - int(time.time() * 1000))

    def _poll_timeout(self) -> int:
        millis_to_wait_for_connect_timeout = self._millis_to_wait_for_connect_timeout()
        if logger.isEnabledFor(logging.DEBUG):
            if millis_to_wait_for_connect_timeout < 0:
                timeout_millis = DEBUG_INTERVAL_MILLIS
            else:
                timeout_millis = min(DEBUG_INTERVAL_MILLIS, millis_to_wait_for_connect_timeout)
        else:
            timeout_millis = millis_to_wait_for_connect_timeout

        if timeout_millis < 0:
            return None
        return max(1, int(timeout_millis / 1000))

    def _poll(self) -> List[Tuple[selectors.SelectorKey, int]]:
        timeout = self._poll_timeout()

        if logger.isEnabledFor(logging.DEBUG):
            if timeout is None:
                logger.debug("poll timeout: infinity")
            else:
                logger.debug("poll timeout: %ss", timeout)

        events = self._selector.select(timeout)
        if not events and logger.isEnabledFor(logging.DEBUG):  # poll is interrupted by timeout
            self._show_debug_info()
        if events and logger.isEnabledFor(logging.DEBUG):
            logger.debug("events polled: %s", self._events_to_str(events))
        return events

    def _process_connection_timeout(self):
        if not self._connect_timeout_due_millis:
            return
        current = int(time.time() * 1000)
        if logger.isEnabledFor(logging.DEBUG):
            due_diff = {k: f"{max(0, v - current)}ms" for k, v in self._connect_timeout_due_millis.items()}
            logger.debug("checking connection timeout, countdowns: %s", due_diff)
        to_delete = []
        for fd, due_millis in dict(self._connect_timeout_due_millis).items():
            if due_millis <= current:
                channel = self._channels.get(fd)
                if channel and not channel._ever_active:
                    logger.error(f"connection timeout: {channel}")
                    self._close_channel_internally(channel, reason='connect timeout')
                to_delete.append(fd)
        for fd in to_delete:
            self._connect_timeout_due_millis.pop(fd, None)

    def _check_channel_active(self, channel: AbstractChannel):
        if channel._ever_active:
            return
        try:
            channel.socket().getpeername()
            no_peer = False
        except Exception:
            no_peer = True
        if no_peer:
            if not channel._ever_active:
                channel.set_active(False, 'no peer')
                channel.channel_future().set(channel)
            else:               # half close
                pass
            return
        if not channel._ever_active:  # first time to be active
            channel.set_active(True, 'first time to be active')
            channel.channel_future().set(channel)
            self._connect_timeout_due_millis.pop(channel.fileno(), None)

    @log(logger)
    def _start(self):
        self._thread = threading.current_thread()
        self._start_barrier.set()
        logger.debug(f"eventloop (EventFD:{hex(id(self._eventfd))}) started in thread: {self._thread.name}")
        while True:
            if self._stop_polling:
                self._selector.close()
                logger.debug(f"eventloop (EventFD:{hex(id(self._eventfd))}) closed in thread: {self._thread.name}")
                return

            for key, event in self._poll():
                fileno = key.fd
                if fileno == self._eventfd.fileno():  # just to wake up from epoll
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("EventFD %s interrupted", hex(id(self._eventfd)))
                    self._eventfd.unsafe_read()
                    self._eventfd_read_count += 1
                    continue

                channel = self._channels.get(fileno)
                if not channel:
                    logger.debug("channel not found by fileno: %s", fileno)
                    continue

                if channel.is_server():
                    server_channel = channel
                    if not server_channel.is_active():
                        server_channel.set_active(True, reason='server channel is always active')
                    for client_sock, client_addr in server_channel.acceptall():
                        self._total_accepted += 1
                        logger.debug("accepted: %s, address: %s", sockinfo(client_sock), client_addr)
                        server_channel.handler_context().fire_channel_read(client_sock)
                    continue

                if event & selectors.EVENT_WRITE:
                    self._check_channel_active(channel)
                    if not channel.has_pendings():  # has no pending chunks
                        channel.remove_flag(selectors.EVENT_WRITE)
                    else:
                        chunks = channel.pendings()
                        while True:
                            head, *tail = chunks
                            if head.close:  # denote to close locally
                                logger.debug("process chunk with close indicator: %s", channel)
                                self._close_channel_internally(channel, 'chunk with close indicator')
                                break
                            l0 = len(head.buffer)
                            head.buffer = channel.try_send(head.buffer)
                            sent_bytes = l0 - len(head.buffer)
                            self._total_sent += sent_bytes
                            channel._pending_bytes -= sent_bytes
                            if not head.buffer:  # all data sent for this chunk
                                chunks = tail
                                head.future.set_result(True)
                                if not chunks:  # no chunks left
                                    break
                            else:   # still has data to send later for this chunk
                                break
                        channel.set_pendings(chunks)
                        if not channel.has_pendings():
                            channel.remove_flag(selectors.EVENT_WRITE)
                    channel._check_writability()

                if event & selectors.EVENT_READ and fileno in self._channels:
                    buffer, eof = channel.recvall()
                    self._total_received += len(buffer)
                    if buffer:
                        self._check_channel_active(channel)
                        # logger.info("receive: %s bytes: %s", len(buffer), buffer.decode('utf-8').replace('\n', '\\n'))
                        channel.handler_context().fire_channel_read(buffer)
                    elif eof:
                        self._close_channel_internally(channel, 'EOF')
                        continue

            self._process_task_queue()
            self._process_connection_timeout()

    def start(self):
        if self._start_barrier.is_set():
            return
        with self._lock:
            if self._start_barrier.is_set():
                return
            self._pool.submit(self._start)
            self._start_barrier.wait()


@define(slots=False)
class EventLoopGroup:

    num: int = field(default=1)                # 1 is enough for most cases, especially for high IO
    prefix: str = field(default="")  # prefix for eventloop name

    def __attrs_post_init__(self):
        self.pool = create_thread_pool(self.num, self.prefix)
        self.eventloops = [EventLoop(self.pool) for _ in range(self.pool._max_workers)]
        self._iter = itertools.cycle(self.eventloops)
        pass

    def get_eventloop(self) -> EventLoop:
        return self._iter.__next__()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for eventloop in self.eventloops:
            eventloop.stop()
        self.pool.shutdown()



import sys
import logging
import itertools
import traceback
from functools import wraps
from concurrent import futures
import selectors

_counter = itertools.count()


def sockinfo(sock):
    """Example: [id: 0xd829bade, L:/127.0.0.1:2069 - R:/127.0.0.1:55666]"""
    sock_id = hex(id(sock))
    fileno = sock.fileno()
    s_addr = None
    try:
        s_addr, s_port = sock.getsockname()
        d_addr, d_port = sock.getpeername()
        return f"[id: {sock_id}, fd: {fileno}, L:/{s_addr}:{s_port} - R:/{d_addr}:{d_port}]"
    except Exception:
        if s_addr:
            return f"[id: {sock_id}, fd: {fileno}, L:/{s_addr}:{s_port}]"
        else:
            # return f"[id: {sock_id}, fd: {fileno}, CLOSED]"
            return str(sock)


def create_thread_pool(n=None, prefix=''):
    return futures.ThreadPoolExecutor(max_workers=n, thread_name_prefix=prefix or f'EventloopGroup-{next(_counter)}')


def log(logger: logging.Logger = None, console: bool = True):

    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kw):
            try:
                return func(*args, **kw)
            except Exception as e:
                if logger:
                    logger.exception('unhandled exception: %s', str(e))
                if console:
                    print(traceback.format_exc(), file=sys.stderr)
        return wrapper
    return decorate


class LoggerAdapter(logging.LoggerAdapter):
    def __init__(self, logger, prefix='py-netty'):
        super(LoggerAdapter, self).__init__(logger, {})
        self.prefix = prefix

    def process(self, msg, kwargs):
        if self.prefix:
            return '[%s] %s' % (self.prefix, msg), kwargs
        else:
            return msg, kwargs


def flag_to_str(flag):
    flags = []
    if flag & selectors.EVENT_READ:
        flags.append("R")
    if flag & selectors.EVENT_WRITE:
        flags.append("W")
    return "|".join(flags)



from attrs import define, field
from concurrent.futures import Future

EMPTY_BUFFER = b''


@define(slots=True)
class Chunk:

    buffer: bytes = field()
    future: Future = field(default=None)
    close: bool = field(default=False)

    def __attrs_post_init__(self):
        self.future = self.future or Future()

    def __str__(self):
        return f"Chunk(bytes={len(self.buffer)}, future={self.future}, close={self.close})"

    def __repr__(self):
        return self.__str__()



import os
import ssl
import time
import socket
import logging
from functools import wraps
from concurrent.futures import Future
from typing import Callable, List, Union, Tuple, Optional
from .bytebuf import Chunk, EMPTY_BUFFER
from .handler import LoggingChannelHandler
from .utils import sockinfo, log, LoggerAdapter, flag_to_str
import selectors
from attrs import define, field
import errno

logger = LoggerAdapter(logging.getLogger(__name__))


_DEFAULT_LOW_WATER_MARK = int(os.getenv('PY_NETTY_LOW_WATER_MARK', 32 * 1024))
_DEFAULT_HIGH_WATER_MARK = int(os.getenv('PY_NETTY_HIGH_WATER_MARK', 64 * 1024))
_ROUNDS = int(os.getenv('PY_NETTY_TUNING_ROUNDS', 16))
_INITIAL_BUFFER_SIZE = int(os.getenv('PY_NETTY_TUNING_INIT_BUFFER_SIZE', 1024))
_MIN_BUFFER_SIZE = int(os.getenv('PY_NETTY_TUNING_MIN_BUFFER_SIZE', _INITIAL_BUFFER_SIZE >> 1))
_MAX_BUFFER_SIZE = int(os.getenv('PY_NETTY_TUNING_MAX_BUFFER_SIZE', _INITIAL_BUFFER_SIZE << 4))


@define(slots=True, kw_only=True)
class ChannelInfo:

    sock: socket.socket = field()
    id: str = field()
    sockname: Tuple[str, int] = field()
    peername: Tuple[str, int] = field()
    fileno: int = field(default=-1)

    @classmethod
    def of(cls, sock: socket.socket):
        return cls(
            sock=sock,
            id=hex(id(sock)),
            sockname=sock.getsockname()[:2],
            peername=sock.getpeername()[:2],
            fileno=sock.fileno()
        )


def adaptive_bufsize(previous_bufsize, data_size):
    if data_size < (previous_bufsize >> 1) and previous_bufsize > _MIN_BUFFER_SIZE:
        return max(previous_bufsize >> 1, _MIN_BUFFER_SIZE)
    elif data_size == previous_bufsize and previous_bufsize < _MAX_BUFFER_SIZE:
        return min(previous_bufsize << 1, _MAX_BUFFER_SIZE)
    else:
        return previous_bufsize


@define(slots=True)
class AbstractChannel:

    _eventloop: 'EventLoop' = field()
    _socket: socket.socket = field()
    _handler_initializer: Callable = field(factory=LoggingChannelHandler)

    def __attrs_post_init__(self):
        self._fileno = self._socket.fileno()
        assert self._fileno > 0
        self._close_future = ChannelFuture(self)
        self._active = False
        self._handler = None    # lazy initialization
        self._flag = 0          # interested events
        self._server_channel = False
        self._ever_active = False
        self._sockinfo = None
        self._channelinfo = None
        self._channel_future = ChannelFuture(self)

    def channel_future(self) -> 'ChannelFuture':
        return self._channel_future

    def eventloop(self) -> 'EventLoop':
        return self._eventloop

    def id(self):
        return str(hex(id(self.socket())))

    def context(self) -> 'ChannelContext':
        return ChannelContext(self)

    def set_flag(self, flag):
        self._flag = flag

    def socket(self) -> socket.socket:
        return self._socket

    def channelinfo(self) -> Optional[ChannelInfo]:
        """Include ORIGINAL socket info, even if the sock is closed."""
        if self._channelinfo is None:
            _ = str(self)
        return self._channelinfo

    def register(self) -> 'ChannelFuture':
        return self.eventloop().register(self)

    def unregister(self) -> 'ChannelFuture':
        return self.eventloop().unregister(self)

    def flag(self):
        return self._flag

    def add_flag(self, flag):
        if not self.in_eventloop():
            self._eventloop.submit_task(lambda: self.add_flag(flag))
            return

        if self._flag & flag:
            return

        self._flag |= flag
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("add flag %s(%s) to channel %s/%s, current flag: %s(%s)",
                         flag, flag_to_str(flag), self.id(), self._fileno, self._flag, flag_to_str(self._flag))
        try:
            self.eventloop().modify_flag(self)
        except Exception:
            logger.exception("add flag %s(%s) to channel %s/%s failed (maybe fileno is closed)", flag, flag_to_str(flag), self.id(), self._fileno)

    def remove_flag(self, flag):
        if not self.in_eventloop():
            self._eventloop.submit_task(lambda: self.remove_flag(flag))
            return

        if not self._flag & flag:
            return

        self._flag &= ~flag

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("remove flag %s(%s) from channel %s/%s, current flag: %s(%s)",
                         flag, flag_to_str(flag), self.id(), self._fileno, self._flag, flag_to_str(self._flag))
        try:
            self.eventloop().modify_flag(self)
        except Exception:       # maybe fileno is closed
            logger.exception("remove flag %s(%s) from channel %s failed", flag, flag_to_str(flag), self.id())

    def handler(self):
        if self._handler is None:
            self._handler = self._handler_initializer()
        return self._handler

    def handler_context(self) -> 'ChannelHandlerContext':
        return ChannelHandlerContext(self)

    def is_server(self):
        """Returns True if this channel is related to a server listening socket, False otherwise."""
        return self._server_channel

    def fileno(self):
        return self._socket.fileno()

    def fileno0(self) -> int:
        """origin fileno"""
        return self._fileno

    def close_future(self) -> 'ChannelFuture':
        return self._close_future

    def is_active(self):
        if self.close_future().done():
            return False
        return self._active

    def _refresh_sock_info(self) -> str:
        self._sockinfo = None
        self._channelinfo = None
        return str(self)

    @log(logger)
    def set_active(self, active, reason=''):
        origin = self._active
        if origin != active:
            logger.debug("set channel %s active status: %s, reason: %s", self.id(), active, reason)
        self._active = active
        if origin is True and active is False:
            self.handler_context().fire_channel_inactive()
        if origin is False and active is True:
            self._ever_active = True
            self._refresh_sock_info()
            if isinstance(self.socket(), ssl.SSLSocket):
                try:
                    s = time.perf_counter()
                    self.socket().do_handshake(True)
                    cost = time.perf_counter() - s  # seconds
                    if cost > 1:
                        logger.warning("ssl handshake cost: ~%ss", round(cost, 2))
                except socket.timeout:
                    logger.exception("ssl handshake timeout")
                    self.close(True)
                except socket.error as socket_err:
                    logger.debug("ssl handshake error: %s", str(socket_err))
                    # if errno.ENOTCONN is not socket_err.errno:
                    #     logger.debug("ssl handshake error: %s", str(socket_err))
                else:
                    self.handler_context().fire_channel_handshake_complete()
                # finally:
                #     with suppress(Exception):
                #         self.socket().settimeout(0)

            self.handler_context().fire_channel_active()

    def close(self, force=False):
        if force:
            self.close_forcibly()
        else:                   # gracefully
            self.close_gracefully()

    def close_forcibly(self) -> 'ChannelFuture':
        if not self.in_eventloop():
            self._eventloop.submit_task(self.close_forcibly)
            return self.close_future()
        logger.debug(f"Closing channel FORCIBLY: {self}")
        self.eventloop()._close_channel_internally(self, 'close channel forcibly')
        return self.close_future()

    def close_gracefully(self) -> 'ChannelFuture':
        if not self.in_eventloop():
            self._eventloop.submit_task(self.close_gracefully)
            return self.close_future()

        logger.debug(f"Closing channel (active:{self.is_active()}) GRACEFULLY: {self}")
        if not self.is_active():
            return self.close_future()

        if self.is_server():
            self.eventloop()._close_channel_internally(self, 'close server channel gracefully')
        else:                  # client channel
            self.add_pending(Chunk(EMPTY_BUFFER, self.close_future().future, True))
        return self.close_future()

    def in_eventloop(self):
        return self._eventloop.in_eventloop()

    def __str__(self):
        if not self._sockinfo:
            self._sockinfo = sockinfo(self._socket)
        if not self._channelinfo:
            try:
                self._channelinfo = ChannelInfo.of(self._socket)
            except Exception:
                pass
        if self._channelinfo:
            ci = self._channelinfo
            l_addr = ':'.join(map(str, ci.sockname))
            r_addr = ':'.join(map(str, ci.peername))
            sign = '-'
            if not self._ever_active:
                sign = '?'
            elif not self.is_active():
                sign = '!'
            return f"[id:{ci.id}, fd:{ci.fileno}, L:/{l_addr} {sign} R:/{r_addr}]"
        if not self._ever_active:
            return self._sockinfo.replace('-', '?')
        if not self.is_active():
            return self._sockinfo.replace('-', '!')
        return self._sockinfo


class NioSocketChannel(AbstractChannel):

    def __init__(self, eventloop: 'EventLoop', sock: socket.socket, handler_initializer: Callable, connect_timeout_millis: int = 3000):
        super().__init__(eventloop, sock, handler_initializer)
        self._pendings = []     # [Chunk, ...]
        self._pending_bytes = 0
        self._writable = True
        self._auto_read = True
        try:
            self.socket().setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        except Exception:
            # (Windows) OSError: [WinError 10022] An invalid argument was supplied
            pass
        self._connect_timeout_millis = connect_timeout_millis

    def is_writable(self) -> bool:
        return self._writable

    def _check_writability(self):
        # assert self.in_eventloop()
        writable = self._writable
        if self._pending_bytes >= _DEFAULT_HIGH_WATER_MARK:
            writable = False
        elif self._pending_bytes <= _DEFAULT_LOW_WATER_MARK:
            writable = True
        if self._writable != writable:
            self._writable = writable
            self.handler_context().fire_channel_writability_changed()

    def connect_timeout_millis(self) -> int:
        return self._connect_timeout_millis

    def pendings(self) -> List['Chunk']:
        return self._pendings

    def set_pendings(self, pendings: List['Chunk']):
        self._pendings = pendings

    def add_pending(self, chunk: 'Chunk'):
        if chunk is None:
            return
        if chunk.close is False and not chunk.buffer:
            return
        self._pendings.append(chunk)
        self._pending_bytes += len(chunk.buffer)
        self.add_flag(selectors.EVENT_WRITE)

    def is_auto_read(self) -> bool:
        return self._auto_read

    # a way of flow control
    def set_auto_read(self, auto_read: bool):
        if self._auto_read == auto_read:
            return
        self._auto_read = auto_read
        if auto_read:
            self.add_flag(selectors.EVENT_READ)
        else:
            self.remove_flag(selectors.EVENT_READ)

    def has_pendings(self) -> bool:
        return len(self._pendings) > 0

    def write(self, buffer, channel_future: 'ChannelFuture' = None) -> 'ChannelFuture':
        cf = channel_future or ChannelFuture(self)
        if not self.in_eventloop():
            self._eventloop.submit_task(lambda: self.write(buffer, cf))
            return cf
        self.add_pending(Chunk(buffer, cf.future))
        return cf

    def try_send(self, bytebuf: bytes, spin=1) -> bytes:
        if not bytebuf:
            return b''
        total_sent = 0
        while total_sent < len(bytebuf):
            try:
                total_sent += self.socket().send(bytebuf[total_sent:])
            except socket.error as socket_err:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("try_send socket.error: %s, spin: %s", str(socket_err), spin)
                if spin > 0:
                    spin -= 1
                    continue
                break
        return bytebuf[total_sent:]

    def recvall(self) -> (bytes, bool):
        # if isinstance(self.socket(), ssl.SSLSocket):
        #     return self.recvall_ssl()
        buffer = b''
        bufsize = _INITIAL_BUFFER_SIZE
        rounds = 0
        total_costs = 0
        while True:
            rounds += 1
            try:
                s = time.perf_counter()
                received = self.socket().recv(bufsize)
                cost = time.perf_counter() - s  # seconds
                total_costs += cost
                if not received:  # EOF
                    return buffer, True
                recv_len = len(received)
                bufsize = adaptive_bufsize(bufsize, recv_len)
                buffer += received
                if rounds == _ROUNDS or total_costs > 0.1 or cost > 0.01:
                    if logger.isEnabledFor(logging.DEBUG):
                        cost = round(cost, 3)
                        total_costs = round(total_costs, 3)
                        sock_info = sockinfo(self.socket())
                        logger.debug(f"yield from recvall, rounds:{rounds}, current cost:{cost * 1000}ms, total cost:{total_costs * 1000}ms, bufsize:{bufsize}, buffer: {len(buffer)}, socket:{sock_info}")
                    return buffer, False
            # except ssl.SSLWantReadError:  # for ssl socket
            #     logger.debug("recvall ssl.SSLWantReadError, readable: %s", self.is_readable())
            #     if self.is_readable():
            #         continue
            #     return buffer, False
            except socket.error as socket_err:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("recvall socket.error: %s, readable: %s", str(socket_err), self.is_readable())
                if self.is_readable():
                    continue
                return buffer, socket_err.errno in (errno.ECONNABORTED, errno.ECONNRESET)

    def is_readable(self) -> bool:
        try:
            return not self.socket().recv(16, socket.MSG_DONTWAIT | socket.MSG_PEEK) == b''
        except Exception:
            return False

    # def recvall_ssl(self) -> (bytes, bool):
    #     buffer = b''
    #     while True:
    #         try:
    #             received = self.socket().recv(1024)
    #             buffer += received
    #             if not received:  # EOF
    #                 return buffer, True
    #         except ssl.SSLWantReadError:
    #             if self.is_readable():
    #                 continue
    #             return buffer, False
    #         except socket.error:
    #             return buffer, False


class NioServerSocketChannel(AbstractChannel):

    def __init__(self, eventloop: 'EventLoop', sock: socket.socket, handler_initializer: Callable):
        super().__init__(eventloop, sock, handler_initializer)
        self._server_channel = True

    def acceptall(self) -> list:  # [(socket, address), ...]
        result = []
        while True:
            try:
                result.append(self.socket().accept())
            except socket.error:
                return result


@define(slots=True)
class ChannelContext:
    _channel: AbstractChannel = field()

    def close(self):
        self._channel.close()

    def write(self, buffer):
        self._channel.write(buffer)

    def channel(self):
        return self._channel


# annotation
def _catch_exception(func):
    @wraps(func)
    def inner(self, *args, **kwargs):
        try:
            func(self, *args, **kwargs)
        except Exception as e:
            self.fire_exception_caught(e)
    return inner


@define(slots=True)
class ChannelHandlerContext:

    _channel: AbstractChannel = field()

    def close(self):
        self._channel.close()

    def write(self, bytebuf) -> 'ChannelFuture':
        return self._channel.write(bytebuf)

    def channel(self):
        return self._channel

    def handler(self):
        return self._channel.handler()

    def fire_exception_caught(self, exception):
        try:
            self.handler().exception_caught(self, exception)
        except Exception:
            logger.exception(f"Exception caught while handling exception: {exception}")

    @_catch_exception
    def fire_channel_registered(self):
        self.handler().channel_registered(self)

    @_catch_exception
    def fire_channel_unregistered(self):
        self.handler().channel_unregistered(self)

    @_catch_exception
    def fire_channel_read(self, msg: Union[bytes, socket.socket]):
        self.handler().channel_read(self, msg)

    @_catch_exception
    def fire_channel_active(self):
        self.handler().channel_active(self)

    @_catch_exception
    def fire_channel_inactive(self):
        self.handler().channel_inactive(self)

    @_catch_exception
    def fire_channel_writability_changed(self):
        self.handler().channel_writability_changed(self)

    @_catch_exception
    def fire_channel_handshake_complete(self):
        self.handler().channel_handshake_complete(self)


@define(slots=True)
class ChannelFuture:

    _channel: AbstractChannel = field()
    future: Future = field(default=None)

    def __attrs_post_init__(self):
        self.future = self.future or Future()

    def channel(self) -> AbstractChannel:
        return self._channel

    def close_future(self) -> 'ChannelFuture':
        return self.channel().close_future()

    def sync(self) -> 'ChannelFuture':
        self.future.result()
        return self

    def done(self) -> bool:
        return self.future.done()

    def set(self, channel: AbstractChannel) -> None:
        if self.future.done():
            return
        self.future.set_result(channel)

    def add_listener(self, listener: Callable) -> None:
        self.future.add_done_callback(lambda f: listener(self))



from .bootstrap import Bootstrap, ServerBootstrap
from .handler import ChannelHandlerAdapter
from .eventloop import EventLoopGroup

__all__ = ['Bootstrap', 'ServerBootstrap', 'ChannelHandlerAdapter', 'EventLoopGroup']



import ssl
import typing
import socket
import logging
from functools import lru_cache
from .eventloop import EventLoopGroup
from .handler import EchoChannelHandler, ChannelHandlerAdapter
from .channel import ChannelFuture, ChannelContext, NioSocketChannel, NioServerSocketChannel
from attrs import define, field


logger = logging.getLogger(__name__)


def _handler_initializer():
    return EchoChannelHandler()


@lru_cache(maxsize=8)
def _client_ssl_context(verify=True):
    if verify:
        return ssl.create_default_context()
    else:                       # no verify
        ssl_context = ssl._create_unverified_context()
        ssl_context.minimum_version = ssl.TLSVersion.TLSv1_2
        ssl_context.set_ciphers("ALL")
        return ssl_context


@lru_cache(maxsize=8)
def _server_ssl_context(certfile, keyfile):
    s_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    s_context.load_cert_chain(certfile, keyfile)
    return s_context


@define(slots=True)
class Bootstrap:
    eventloop_group: EventLoopGroup = field(factory=EventLoopGroup)
    handler_initializer: typing.Callable = field(default=_handler_initializer)
    tls: bool = False
    verify: bool = True
    ssl_context_cb: typing.Callable = None

    def _create_ssl_context(self):
        ctx = _client_ssl_context(self.verify)
        if self.ssl_context_cb:
            try:
                self.ssl_context_cb(ctx)
            except Exception as e:
                logger.error("Error in ssl_context_cb(client): %s", e)
        return ctx

    def _wrap_ssl_socket(self, sock, server_hostname_or_address):
        return self._create_ssl_context().wrap_socket(sock, server_hostname=server_hostname_or_address)

    def connect(self, address, port, ensure_connected: bool = False) -> ChannelFuture:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # if ensure_connected or self.tls:
        if ensure_connected:
            sock.connect((address, port))
            if self.tls:
                sock = self._wrap_ssl_socket(sock, address)
            sock.setblocking(False)
        else:
            sock.setblocking(False)
            if self.tls:
                sock = self._wrap_ssl_socket(sock, address)
            sock.connect_ex((address, port))  # non blocking
        return NioSocketChannel(
            self.eventloop_group.get_eventloop(),
            sock,
            handler_initializer=self.handler_initializer
        ).register()


@define(slots=True)
class ServerBootstrap:
    parant_group: EventLoopGroup = field(factory=EventLoopGroup)
    child_group: EventLoopGroup = field(factory=EventLoopGroup)
    child_handler_initializer: typing.Callable = field(default=_handler_initializer)
    certfile: str = None
    keyfile: str = None
    ssl_context_cb: typing.Callable = None

    def bind(self, address='localhost', port=-1) -> ChannelFuture:
        assert port > 0
        assert ((self.certfile is not None) ^ (self.keyfile is not None)) is False, "Both certfile and keyfile must be specified"
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if self.certfile and self.keyfile:
            ssl_ctx = _server_ssl_context(self.certfile, self.keyfile)
            if self.ssl_context_cb:
                try:
                    self.ssl_context_cb(ssl_ctx)
                except Exception as e:
                    logger.error("Error in ssl_context_cb(server): %s", e)
            server_socket = ssl_ctx.wrap_socket(server_socket, server_side=True)
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server_socket.bind((address, port))
        server_socket.listen(128)
        server_socket.setblocking(0)
        eventloop = self.parant_group.get_eventloop()

        class _ChannelInitializer(ChannelHandlerAdapter):
            def channel_read(this, ctx: ChannelContext, client_socket: socket.socket):
                logger.debug("Initializing client socket: %s", client_socket)
                client_socket.setblocking(0)
                NioSocketChannel(
                    self.child_group.get_eventloop(),
                    client_socket,
                    handler_initializer=self.child_handler_initializer
                ).register()

        return NioServerSocketChannel(eventloop, server_socket, handler_initializer=_ChannelInitializer).register()
        # return eventloop.register(server_socket, is_server=True, handler_initializer=_ChannelInitializer)
